@@ -17,13 +17,18 @@ object SpinalConfig extends spinal.core.SpinalConfig(
 ){
   //Insert a compilation phase which will add a  (* ram_style = "block" *) on all synchronous rams.
   phasesInserters += {(array) => array.insert(array.indexWhere(_.isInstanceOf[PhaseAllocateNames]) + 1, new ForceRamBlockPhase)}
+  phasesInserters += {(array) => array.insert(array.indexWhere(_.isInstanceOf[PhaseAllocateNames]) + 1, new NoRwCheckPhase)}
 }
 
 case class ArgConfig(
   debug : Boolean = false,
+  hardwareBreakpointCount : Int = 0,
   iCacheSize : Int = 4096,
   dCacheSize : Int = 4096,
+  pmpRegions : Int = 0,
+  pmpGranularity : Int = 256,
   mulDiv : Boolean = true,
+  cfu : Boolean = false,
   atomics: Boolean = false,
   compressedGen: Boolean = false,
   singleCycleMulDiv : Boolean = true,
@@ -55,12 +60,16 @@ object GenCoreDefault{
     val parser = new scopt.OptionParser[ArgConfig]("VexRiscvGen") {
       //  ex :-d    or   --debug
       opt[Unit]('d', "debug")    action { (_, c) => c.copy(debug = true)   } text("Enable debug")
+      opt[Int]("hardwareBreakpointCount")     action { (v, c) => c.copy(hardwareBreakpointCount = v) } text("Set number of available hardware breakpoints, defaults to 0")
       // ex : -iCacheSize=XXX
       opt[Int]("iCacheSize")     action { (v, c) => c.copy(iCacheSize = v) } text("Set instruction cache size, 0 mean no cache")
       // ex : -dCacheSize=XXX
       opt[Int]("dCacheSize")     action { (v, c) => c.copy(dCacheSize = v) } text("Set data cache size, 0 mean no cache")
+      opt[Int]("pmpRegions")    action { (v, c) => c.copy(pmpRegions = v)   } text("Number of PMP regions, 0 disables PMP")
+      opt[Int]("pmpGranularity")    action { (v, c) => c.copy(pmpGranularity = v)   } text("Granularity of PMP regions (in bytes)")
       opt[Boolean]("mulDiv")    action { (v, c) => c.copy(mulDiv = v)   } text("set RV32IM")
-      opt[Boolean]("atomics")    action { (v, c) => c.copy(mulDiv = v)   } text("set RV32I[A]")
+      opt[Boolean]("cfu")       action { (v, c) => c.copy(cfu = v)   } text("If true, add custom function unit interface")
+      opt[Boolean]("atomics")    action { (v, c) => c.copy(atomics = v)   } text("set RV32I[A]")
       opt[Boolean]("compressedGen")    action { (v, c) => c.copy(compressedGen = v)   } text("set RV32I[C]")
       opt[Boolean]("singleCycleMulDiv")    action { (v, c) => c.copy(singleCycleMulDiv = v)   } text("If true, MUL/DIV are single-cycle")
       opt[Boolean]("singleCycleShift")    action { (v, c) => c.copy(singleCycleShift = v)   } text("If true, SHIFTS are single-cycle")
@@ -90,7 +99,7 @@ object GenCoreDefault{
             cmdForkOnSecondStage = false,
             cmdForkPersistence = false, //Not required as the wishbone bridge ensure it
             compressedGen = argConfig.compressedGen,
-            memoryTranslatorPortConfig = if(linux) MmuPortConfig(portTlbSize = 4)
+            memoryTranslatorPortConfig = if(linux) MmuPortConfig(portTlbSize = 4) else null
           )
         }else {
           new IBusCachedPlugin(
@@ -98,11 +107,11 @@ object GenCoreDefault{
             relaxedPcCalculation = argConfig.relaxedPcCalculation,
             prediction = argConfig.prediction,
             compressedGen = argConfig.compressedGen,
-            memoryTranslatorPortConfig = if(linux) MmuPortConfig(portTlbSize = 4),
+            memoryTranslatorPortConfig = if(linux) MmuPortConfig(portTlbSize = 4) else null,
             config = InstructionCacheConfig(
               cacheSize = argConfig.iCacheSize,
               bytePerLine = 32,
-              wayCount = 1,
+              wayCount = if(linux) ((argConfig.iCacheSize + 4095) / 4096) else 1,
               addressWidth = 32,
               cpuDataWidth = 32,
               memDataWidth = 32,
@@ -120,7 +129,7 @@ object GenCoreDefault{
             catchAddressMisaligned = true,
             catchAccessFault = true,
             withLrSc = linux || argConfig.atomics,
-            memoryTranslatorPortConfig = if(linux) MmuPortConfig(portTlbSize = 4)
+            memoryTranslatorPortConfig = if(linux) MmuPortConfig(portTlbSize = 4) else null
           )
         }else {
           new DBusCachedPlugin(
@@ -131,7 +140,7 @@ object GenCoreDefault{
             config = new DataCacheConfig(
               cacheSize = argConfig.dCacheSize,
               bytePerLine = 32,
-              wayCount = 1,
+              wayCount = if(linux) ((argConfig.dCacheSize + 4095) / 4096) else 1,
               addressWidth = 32,
               cpuDataWidth = 32,
               memDataWidth = 32,
@@ -139,18 +148,21 @@ object GenCoreDefault{
               catchIllegal = true,
               catchUnaligned = true,
               withLrSc = linux || argConfig.atomics,
-              withAmo = linux,
+              withAmo = linux || argConfig.atomics,
               earlyWaysHits = argConfig.dBusCachedEarlyWaysHits
             ),
-            memoryTranslatorPortConfig = if(linux) MmuPortConfig(portTlbSize = 4),
+            memoryTranslatorPortConfig = if(linux) MmuPortConfig(portTlbSize = 4) else null,
             csrInfo = true
           )
         },
-        if(linux) new MmuPlugin(
+        if (linux) new MmuPlugin(
           ioRange = (x => x(31 downto 28) === 0xB || x(31 downto 28) === 0xE || x(31 downto 28) === 0xF)
-        )  else new StaticMemoryTranslatorPlugin(
+        ) else if (argConfig.pmpRegions > 0) new PmpPlugin(
+          regions = argConfig.pmpRegions, granularity = argConfig.pmpGranularity, ioRange = _.msb
+        ) else new StaticMemoryTranslatorPlugin(
           ioRange      = _.msb
         ),
+
         new DecoderSimplePlugin(
           catchIllegalInstruction = true
         ),
@@ -178,7 +190,8 @@ object GenCoreDefault{
           pessimisticAddressMatch = false
         ),
         new BranchPlugin(
-          earlyBranch = false,
+          // If using CFU, use earlyBranch to avoid incorrect CFU execution
+          earlyBranch = argConfig.cfu,
           catchAddressMisaligned = true
         ),
         new CsrPlugin(
@@ -187,6 +200,7 @@ object GenCoreDefault{
             case "all" => CsrPluginConfig.all(mtvecInit = argConfig.machineTrapVector)
             case "linux" => CsrPluginConfig.linuxFull(mtVecInit = argConfig.machineTrapVector).copy(ebreakGen = false)
             case "linux-minimal" => CsrPluginConfig.linuxMinimal(mtVecInit = argConfig.machineTrapVector).copy(ebreakGen = false)
+            case "secure" => CsrPluginConfig.secure(argConfig.machineTrapVector)
           }
         ),
         new YamlPlugin(argConfig.outputFile.concat(".yaml"))
@@ -221,7 +235,50 @@ object GenCoreDefault{
 
       // Add in the Debug plugin, if requested
       if(argConfig.debug) {
-        plugins += new DebugPlugin(ClockDomain.current.clone(reset = Bool().setName("debugReset")))
+        plugins += new DebugPlugin(
+          debugClockDomain = ClockDomain.current.clone(reset = Bool().setName("debugReset")),
+          hardwareBreakpointCount = argConfig.hardwareBreakpointCount 
+        )
+      }
+
+      // CFU plugin/port
+      if(argConfig.cfu) {
+        plugins ++= List(
+          new CfuPlugin(
+            stageCount = 1,
+            allowZeroLatency = true,
+            withEnable = false,
+            encodings = List(
+              // CFU R-type
+              CfuPluginEncoding (
+                instruction = M"-------------------------0001011",
+                functionId = List(14 downto 12, 31 downto 25),
+                input2Kind = CfuPlugin.Input2Kind.RS
+              )
+              //,
+              // CFU I-type
+              //CfuPluginEncoding (
+              //  instruction = M"-----------------000-----0101011",
+              //  functionId = List(23 downto 20),
+              //  input2Kind = CfuPlugin.Input2Kind.IMM_I
+              //)
+            ),
+            busParameter = CfuBusParameter(
+              CFU_VERSION = 0,
+              CFU_INTERFACE_ID_W = 0,
+              CFU_FUNCTION_ID_W = 10,
+              CFU_REORDER_ID_W = 0,
+              CFU_REQ_RESP_ID_W = 0,
+              CFU_STATE_INDEX_NUM = 0,
+              CFU_INPUTS = 2,
+              CFU_INPUT_DATA_W = 32,
+              CFU_OUTPUTS = 1,
+              CFU_OUTPUT_DATA_W = 32,
+              CFU_FLOW_REQ_READY_ALWAYS = false,
+              CFU_FLOW_RESP_READY_ALWAYS = false
+            )
+          )
+        )
       }
 
       // CPU configuration
@@ -274,4 +331,25 @@ class ForceRamBlockPhase() extends spinal.core.internals.Phase{
     }
   }
   override def hasNetlistImpact: Boolean = false
-}
\ No newline at end of file
+}
+
+class NoRwCheckPhase() extends spinal.core.internals.Phase{
+  override def impl(pc: PhaseContext): Unit = {
+    pc.walkBaseNodes{
+      case mem: Mem[_] => {
+        var doit = false
+        mem.dlcForeach[MemPortStatement]{
+          case _ : MemReadSync => doit = true
+          case _ =>
+        }
+        mem.dlcForeach[MemPortStatement]{
+          case p : MemReadSync if p.readUnderWrite != dontCare  => doit = false
+          case _ =>
+        }
+        if(doit) mem.addAttribute("no_rw_check")
+      }
+      case _ =>
+    }
+  }
+  override def hasNetlistImpact: Boolean = false
+}
